#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifdef VMS
#include <types.h>
#else
#include <sys/types.h>
#endif
#include <time.h>
#include <Xm/MessageB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Mrm/MrmAppl.h>
#include <X11/X.h>
#include <DXm/DECspecific.h>
#ifdef VMS
#include <DECw$Cursor.h>
#else
#include <X11/decwcursor.h>
#endif
#include "XTS.h"
#include "xts_common.h"
#include "xtsu_time_unix.h"
#include "xtsu_database.h"

#define minfocus 0500      /* fastest focus scanning speed; .5s */
#define maxwin 200
#define maxchg 60          /* maximum different types of charge-to */
#define maxname 57         /* maximum size of charge description field */
#define maxothers 16       /* maximum number of ignored XTS windows */

void VUIT_Read ( char *widget_name );
static int (*xtsu_old_error_handler) ( Display *dpy, XErrorEvent *rep );
static int (*xtsu_old_io_error_handler) ( Display *dpy );

static Cursor cursor;
static Widget w_list, w_window_list;
static int editing; /* entry number being edited */
static int pending; /* new window association pending */
static int blinks; /* number of blink cycles done due to pending active */
static int deleting; /* window list entry selected for deletion */
static char storage[256]; /* file name */
static XtIntervalId i_focus, i_saver, i_update, i_project;

static Window wn_toplevel, wn_others[maxothers];

static struct setup
{
    /* timesheet display */
    int charging;       /* whether real time is being charged */
    int mode;           /* list display mode 0=day, 1=week, 2=total */
    int totals;         /* totals line enable */
    int format;         /* list display format 0=hh:mm, 1=hh:mm:ss, 2=h.hh */
    int update_timer;   /* list display update timer */
    int pause_action;   /* pause action, 0=none, 1=no charge, 2=top */
    int pause_thresh;   /* time before a pause is a pause, in seconds */
    
    /* window focus monitoring */
    int focus_enable;   /* focus monitoring enable */
    int focus_bell;     /* bell on new window focus */
    int focus_timer;    /* focus timer interval */
    int focus_revert;   /* whether to revert to default on unassociated window */
    int focus_delay;    /* delay before first window scan */
    
    /* automatic timesheet save */
    int saver_enable;   /* file save timer enable */
    int saver_timer;    /* file save timer interval */
    
    /* whether to use icon */
    int icon;
    
    /* whether to update .project file */
    int project;
    
    /* whether to purge old files */
    int purge;
    
    int work_week_enable;  /* use work week rollover mode */
    int work_week_start;    /* relative work week start day */

    /* from control file */
    int fiscal_year;
    int fiscal_week;
    int day;               /* work day number */

} setup, will ;

struct counts
{
    int show;
    int reselections;
    struct
    {
        int timer, changes, new, orphans;
    } focus;
} counts;

static int old;               /* index of last known window in */
static struct windows         /* list of known windows */
{
    Window id;                /* key into list; the window id */
    int charge;               /* index of current charge entry */
    /* -1 means window not connected to an entry yet */
    datetime seen;            /* last date time the window was seen; for future list purging */
} window[maxwin] ;
/* ?? this window structure is not initialised at run-time */

static datetime notime;

typedef struct 
{
    int inuse;                /* set to 1 to indicate in-use */
    char name[maxname];
    struct day_x
    {
        datetime total;        /* the accumulated time */
    } d[7] ;
} entry;

static struct charge
{
    datetime previous;            /* previous charge timestamp */
    int now;                  /* currently active charge entry */
    int day;                  /* current day number */
    int selected;             /* index of currently selected entry */
    entry t;                  /* totals */
    entry l[maxchg];
} c;


static void move ( to, from, length )
    char *to, *from;
    int length;
{
    int i;
    
    for (i=0;i<length;i++) to[i] = from[i];
}

char *errno_text ()
/* ??? use strerror() */
{
    switch ( errno )
    {
        case EPERM           : return ( "not owner " );
        case ENOENT          : return ( "no such file or directory " );
        case ESRCH           : return ( "no such process " );
        case EINTR           : return ( "interrupted system call " );
        case EIO             : return ( "i/o error " );
        case ENXIO           : return ( "no such device or address " );
        case E2BIG           : return ( "arg list too long " );
        case ENOEXEC         : return ( "exec format error " );
        case EBADF           : return ( "bad file number " );
        case ECHILD          : return ( "no children " );
#if EAGAIN != EWOULDBLOCK
        case EAGAIN          : return ( "no more processes " );
#endif
        case ENOMEM          : return ( "not enough core " );
        case EACCES          : return ( "permission denied " );
        case EFAULT          : return ( "bad address " );
        case ENOTBLK         : return ( "block device required " );
        case EBUSY           : return ( "mount device busy " );
        case EEXIST          : return ( "file exists " );
        case EXDEV           : return ( "cross-device link " );
        case ENODEV          : return ( "no such device " );
        case ENOTDIR         : return ( "not a directory " );
        case EISDIR          : return ( "is a directory " );
        case EINVAL          : return ( "invalid argument " );
        case ENFILE          : return ( "file table overflow " );
        case EMFILE          : return ( "too many open files " );
        case ENOTTY          : return ( "not a typewriter " );
        case ETXTBSY         : return ( "text file busy " );
        case EFBIG           : return ( "file too large " );
        case ENOSPC          : return ( "no space left on device " );
        case ESPIPE          : return ( "illegal seek " );
        case EROFS           : return ( "read-only file system " );
        case EMLINK          : return ( "too many links " );
        case EPIPE           : return ( "broken pipe " );
        case EDOM            : return ( "math argument " );
        case ERANGE          : return ( "result too large " );
        case EWOULDBLOCK     : return ( "I/O operation would block channel " );
        case EINPROGRESS     : return ( "Operation now in progress " );
        case EALREADY        : return ( "Operation already in progress " );
        case ENOTSOCK        : return ( "Socket operation on non-socket " );
        case EDESTADDRREQ    : return ( "Destination address required " );
        case EMSGSIZE        : return ( "Message too long " );
        case EPROTOTYPE      : return ( "Protocol wrong type for socket " );
        case ENOPROTOOPT     : return ( "Protocol not available " );
        case EPROTONOSUPPORT : return ( "Protocol not supported " );
        case ESOCKTNOSUPPORT : return ( "Socket type not supported " );
        case EOPNOTSUPP      : return ( "Operation not supported on socket " );
        case EPFNOSUPPORT    : return ( "Protocol family not supported " );
        case EAFNOSUPPORT    : return ( "Address family not supported " );
        case EADDRINUSE      : return ( "Address already in use " );
        case EADDRNOTAVAIL   : return ( "Can't assign requested address " );
        case ENETDOWN        : return ( "Network is down " );
        case ENETUNREACH     : return ( "Network is unreachable " );
        case ENETRESET       : return ( "Network dropped connection on reset " );
        case ECONNABORTED    : return ( "Software caused connection abort " );
        case ECONNRESET      : return ( "Connection reset by peer " );
        case ENOBUFS         : return ( "No buffer space available " );
        case EISCONN         : return ( "Socket is already connected " );
        case ENOTCONN        : return ( "Socket is not connected " );
        case ESHUTDOWN       : return ( "Can't send after socket shutdown " );
        case ETOOMANYREFS    : return ( "Too many references: can't splice " );
        case ETIMEDOUT       : return ( "Connection timed out " );
        case ECONNREFUSED    : return ( "Connection refused " );
        case ELOOP           : return ( "Too many levels of symbolic links " );
        case ENAMETOOLONG    : return ( "File name too long " );
        case EHOSTDOWN       : return ( "Host is down " );
        case EHOSTUNREACH    : return ( "No route to host " );
#ifdef VMS
        case EVMSERR         : return ( "VMS-specific error code for non-translatable errors from VMS to C " );
#endif
        default              : return ( "untranslatable error code " );
    }
}

static void xtsu_set_value ( w, resource, value )
    Widget      w;
    char        *resource, *value;
{
    Arg al[1]; 

    XtSetArg ( al[0], resource, value );
    XtSetValues ( w, al, 1 );
}

static void xtsu_get_value ( w, resource, value )
    Widget      w;
    char        *resource, *value;
{
    Arg al[1];
    
    XtSetArg ( al[0], resource, value );
    XtGetValues ( w, al, 1 );
}

Widget xtsu_question_dialog ( char *header, char *text )
{                
    Widget w;
    Arg arglist[10];
    XmString message, title;
    int i;

    message = XmStringCreateSimple ( text );
    title = XmStringCreateSimple ( header );
                
    i = 0;

    XtSetArg ( arglist[i], XmNdialogTitle, title ); i++;
    XtSetArg ( arglist[i], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL ); i++;
    XtSetArg ( arglist[i], XmNmessageString, message );     i++;

    w = XmCreateQuestionDialog ( toplevel_widget, "message", arglist, i );
    
    XmStringFree ( message );
    XmStringFree ( title );
    
    return ( w );
}

/*
**  Issue a self-destroying message dialog box with only an OK button
*/
Widget xtsu_message ( char *text, char *header )
{
    Widget w;
    
    w = xtsu_question_dialog ( text, header );
    XtUnmanageChild ( XmMessageBoxGetChild ( w, XmDIALOG_HELP_BUTTON ) );
    XtUnmanageChild ( XmMessageBoxGetChild ( w, XmDIALOG_CANCEL_BUTTON ) );
    XtAddCallback ( w, XmNokCallback, XtDestroyWidget, 0 );
    XtManageChild ( w );
    return ( w );
}

/*
**  Issue a confirmation dialog box with OK and Cancel buttons.
*/
Widget xtsu_confirm ( char *text, char *header, void (*callback), void *value )
{
    Widget w;
    
    w = xtsu_question_dialog ( text, header );
    XtUnmanageChild ( XmMessageBoxGetChild ( w, XmDIALOG_HELP_BUTTON ) );
    XtAddCallback ( w, XmNcancelCallback, callback, value );
    XtAddCallback ( w, XmNokCallback, callback, value );
    XtManageChild ( w );
    return ( w );
}

static char *xtsu_control_name ()
{
    return xtsu_db_get_string ( "Xts.controlFile",
#ifdef VMS
        "XTS_CONTROL" );
#else
        ".xtscf" );
#endif
}
    
/*
**  Write the control file
*/
xtsu_control_write ()
{
    FILE *cf;
    char *name = xtsu_control_name();

    cf = fopen ( name, "r" );
    if ( cf == NULL )
    {
        cf = fopen ( name, "w" );
    }
    else
    {
        fclose ( cf );
        cf = fopen ( name, "r+" );
        rewind ( cf );
    }
    
    /* write to control file */
    if ( cf == NULL )
    {
        printf ( "xtsu_control_write: can't open control file %s for write\n", name );
    }
    else
    {
        fprintf ( cf, "%d\n", setup.fiscal_year );
        fprintf ( cf, "%d\n", setup.fiscal_week );
        fprintf ( cf, "%d\n", setup.day );
        fclose ( cf );
    }
}

/*
**  Read the control file and return the control file day number
*/
void xtsu_control_read ()
{
    FILE *cf;
    char *name = xtsu_control_name();
    
    /* try to open control file */
    cf = fopen ( name, "r" );
    
    if ( cf == NULL )
    {
        /* open failed, apply defaults */
        setup.fiscal_year  = 1993;  /* ?? need better defaults */
        setup.fiscal_week  = 52;
        setup.day          = c.day;
        
        /* create new file */
        xtsu_control_write ();

        /* issue warning */
        {
            char x[256];
            sprintf ( x, "New control file %s created - using default save file %04d-%02d.xts\n", name, setup.fiscal_year, setup.fiscal_week );
            xtsu_message ( "New Control File", x );
        }
    }
    else
    {
        /* open worked, read data */
        fscanf ( cf, "%d", &setup.fiscal_year  );
        fscanf ( cf, "%d", &setup.fiscal_week  );
        fscanf ( cf, "%d", &setup.day   );
        fclose ( cf );
    }
}            

xtsu_control_increment ()
{
    /* work out the new week and year number */
    setup.fiscal_week = setup.fiscal_week + 1;
    if ( setup.fiscal_week > 52 )
    {
        setup.fiscal_year = setup.fiscal_year + 1;
        setup.fiscal_week = 1;
    }
}

/*
**  Return the work day number
*/
int xtsu_work_day_number()
{
    int i;
    
    i = xtsu_time_daynumber();
    i = ( 7 + i - setup.work_week_start ) % 7;
    return ( i );
}

void xtsu_busy(w)
Widget          w;
{
    /* change the cursor to a wait cursor */
    XDefineCursor ( XtDisplay ( toplevel_widget ), XtWindow ( toplevel_widget ), cursor );
    XmUpdateDisplay ( toplevel_widget );
}

void xtsu_not_busy(w)
Widget          w;
{
    /* restore the cursor */
    XUndefineCursor ( XtDisplay ( toplevel_widget ), XtWindow ( toplevel_widget ) );
    XmUpdateDisplay ( toplevel_widget );
}
    
void xtso_set_icon ( char *string )
{
    xtsu_set_value ( toplevel_widget, XmNiconName, string );
}    
    
/*
**  Perform the .project file update.
**  
**  This is executed by the toolkit as a timeout function.
*/
void xtso_set_project_do ( char *string )
{
    FILE *fp;
    char *home;
    char file[80];

    i_project = 0;  /* timer has expired, reset interval id */
    
#ifdef VMS
    strcpy ( file, "SYS$LOGIN:.PROJECT" );
#else
    home = getenv ( "HOME" );
    if ( home == NULL ) home = "";
    strcpy ( file, home );
    strcat ( file, "/" );
    strcat ( file, ".project" );
#endif
    strcpy ( file, xtsu_db_get_string ( "Xts.projectFile", file ) );
    
    xtsu_busy(toplevel_widget);
    fp = fopen ( file, "w" );
    fputs ( xtsu_db_get_string ( "Xts.projectFilePrefix", " " ), fp );
    fputs ( string, fp );
    fputs ( xtsu_db_get_string ( "Xts.projectFileSuffix", " (xts)" ), fp );
    fputc ( '\n', fp );
    fclose ( fp );
    xtsu_not_busy(toplevel_widget);
}    
    
/*
**  Request change to .project file
**  
**  Queues the actual change to occur in the future.  If another change request
**  appears in the mean time, the original change is cancelled.  This is done
**  so as to avoid impacting the user interface and to reduce the frequency of
**  file system updates.  On OpenVMS, where the file system is not normally
**  write-cached, the .project file is updated ten seconds in the future.
**  On other operating systems, the update is deferred only one second.
*/
void xtso_set_project ( char *string )
{
    int timer;
    static char was[maxname]="";
    
    if ( ! strcmp ( was, string ) ) return;
    strcpy ( was, string );

#ifdef VMS
    timer = 10000;  /* ten seconds */
#else
    timer = 1000;   /* one second */
#endif

    XtRemoveTimeOut(i_project);
    i_project = XtAppAddTimeOut ( app_context, timer, xtso_set_project_do, string );
}

char *xtso_window_to_text ( i )
    int i;
{
    char *n;
    static char l[256];
    int j;
            
    /*
    **  ??
    **  
    **  Annoying buglet here; the window may no longer exist, and so XFetchName
    **  will fail.  Need to verify that the window exists, and if not, blow-away
    **  the window entry; which might require updating the window list display.
    */
    
    n = 0;
    XFetchName(display, window[i].id, &n);
    if (n)
    {
        strcpy ( l, n );
        strcat ( l, "  -->  " );
        XFree(n);
    }
    else
    {
        sprintf ( l, "window %08X  -->  ", window[i].id );
    }

    j = window[i].charge;
    if ( j == -1 )
    {
        strcat ( l, "{not connected}" );
    }
    else
    {
        strcat ( l, c.l[window[i].charge].name );
    }
    
    return ( l );
}

datetime *xtso_text_to_charge_value ( char *text )
{
    switch ( setup.format )
    {
        case 0: return ( xtsu_time_from_hhmm   ( text ));
        case 1: return ( xtsu_time_from_hhmmss ( text ));
        case 2: return ( xtsu_time_from_h_hh   ( text ));
    }
}

char *xtso_charge_value_to_text ( v )
    datetime *v;
{
    switch ( setup.format )
    {
        case 0: return ( xtsu_time_to_hhmm   ( v ));
        case 1: return ( xtsu_time_to_hhmmss ( v ));
        case 2: return ( xtsu_time_to_h_hh   ( v ));
    }
}

char *xtso_charge_line_to_text ( e )
    entry *e;
{
    static char l[256];

    /* list display mode 0=day, 1=week, 2=total */
    switch ( setup.mode )
    {
        /* mode day */
        case 0:
        {
            strcpy ( l, xtso_charge_value_to_text ( &e->d[c.day].total ) );
            strcat ( l, " " );
            break;
        }
        
        /* mode week */
        case 1:
        {
            int j;
        
            strcpy ( l, "" );
            
            for (j=0;j<7;j++)
            {
                strcat ( l, xtso_charge_value_to_text ( &e->d[j].total ) );
                strcat ( l, " " );
            }
            break;
        }
        
        /* mode total */
        case 2:
        {
            int j;
            datetime v;
            
            v = notime;
            for (j=0;j<7;j++) xtsu_time_add ( &e->d[j].total, &v, &v );
            strcpy ( l, xtso_charge_value_to_text ( &v ) );
            strcat ( l, " (t) " );
        }
        
    }
    
    strcat ( l, " " );
    strcat ( l, e->name );
    
    return ( l );
}






/*
**  List Object Functions
*/

/*
**  Clear the list
*/
void xtso_list_clear ( )
{
    XmListDeleteAllItems ( w_list );
}

/*
**  Mark an item as the current chargeable item
**  (causes icon name to be changed if icon changes are enabled)
*/
void xtso_list_mark(n)
    int n;
{
    /* highlight the item */
    XmListSelectPos ( w_list, n+1, FALSE );
    XmListSetBottomPos ( w_list, n+2 );

    /* update icon */
    if ( ( setup.icon ) && ( setup.charging ) )
    {
        char x[maxname+3];
    
        strcpy ( x, xtsu_db_get_string ( "Xts.iconNamePrefix", "(" ) );
        strcat ( x, c.l[n].name );
        strcat ( x, xtsu_db_get_string ( "Xts.iconNameSuffix", ")" ) );
        
        xtso_set_icon ( x );
    }    
    /* update project */
    if ( ( setup.project ) && ( setup.charging ) )
    {
        xtso_set_project ( c.l[n].name );
    }    

}

/*
**  Add a new item to the list
*/
void xtso_list_add ( n )
    int n;
{
    XmString s;
    
    s = XmStringCreateSimple ( xtso_charge_line_to_text ( &c.l[n] ) );
    XmListAddItem ( w_list, s, n+1 );
    XmStringFree ( s );
}

/*
**  Update a specific list item
*/
void xtso_list_update ( n )
    int n;
{
    XmString s;
    
    s = XmStringCreateSimple ( xtso_charge_line_to_text ( &c.l[n] ) );
    XmListReplaceItemsPos ( w_list, &s, 1, n+1 );
    XmStringFree ( s );
}

/*
**  If the totals line is enabled, update the totals and display them.
*/
void xtso_list_totals ( )
{
    int i, j, k;
    XmString s;
    
    /* return if totals not enabled */
    if ( ! setup.totals ) return;
    
    /* count of items seen in list */
    k = 0;
    
    /* initialise the totals per day */
    for (j=0;j<7;j++) c.t.d[j].total = notime;
    
    /* total each entry */
    for (i=0;i<maxchg;i++)
    {
        /* provided it is in use */
        if ( c.l[i].inuse )
        {
            k++;
            for (j=0;j<7;j++)
                xtsu_time_add ( &c.l[i].d[j].total, &c.t.d[j].total, &c.t.d[j].total );
        }
    }

    /* replace the totals line with a new one */
    s = XmStringCreateSimple ( xtso_charge_line_to_text ( &c.t ) );
    XmListReplaceItemsPos ( w_list, &s, 1, k+2 );
    XmStringFree ( s );
}

/*
**  Enable the totals line
*/
void xtso_list_totals_on ( )
{
    XmString s;
    int i, k;
    
    /* count the number of list items */
    k = 0;
    for (i=0;i<maxchg;i++) if ( c.l[i].inuse ) k++;
    if ( k == 0 ) return;

    /* add two blank entries at the end of the list */
    s = XmStringCreateSimple ( " " );
    XmListAddItem ( w_list, s, k+1 );
    XmListAddItem ( w_list, s, k+2 );
    XmStringFree ( s );
    
    /* update the second entry with the current totals */
    xtso_list_totals ( );
}

/*
**  Disable the totals line
*/
void xtso_list_totals_off ( )
{
    int i, k;
    
    /* count the list items */
    k = 0;
    for (i=0;i<maxchg;i++) if ( c.l[i].inuse ) k++;
    if ( k == 0 ) return;
    
    /* delete the totals line(s) */
    XmListDeletePos ( w_list, k+2 );
    XmListDeletePos ( w_list, k+1 );
}


/*
**  Update entire list display
*/
xtso_list_update_all ()
{
    int i;
    for (i=0;i<maxchg;i++) if ( c.l[i].inuse ) xtso_list_update ( i );
    xtso_list_totals ( );
}



/*
**  Charge List
*/

/*
**  Name the charge file
*/
void xtso_charge_name()
{
    if ( setup.work_week_enable )
    {
        sprintf
        (
            storage,
            xtsu_db_get_string ( "Xts.chargeFileNameFormat", "%04d-%02d.xts" ),
            setup.fiscal_year, 
            setup.fiscal_week 
        );
    }
}

/*
**  Save the charge list
*/
static void xtso_charge_save(filename)
    char *filename;
{
    FILE *fp;
    int i, k;
    char tempfile[255];

#ifdef VMS
    /* 
    **  Write over the existing save file.
    **  
    **  Note that for VMS we will end up with a new file.
    */
    fp = fopen ( filename, "w" );
#else
    /*
    **  Invent a temporary output file
    */
    strcpy ( tempfile, "." );
    strcat ( tempfile, filename );
    fp = fopen ( tempfile, "w" );
#endif
    
    /*
    **  If we could not open the file, say so.
    */
    if ( fp == NULL )
    {
        /* issue warning */
        {
            char x[256];
            sprintf ( x, "Save failed.  Could not create output file %s %s", filename, errno_text() );
            xtsu_message ( "Save Failed", x );
        }
        return;
    }

    /* for each entry in the charge list, ... */
    for (i=0;i<maxchg;i++)
    {
        /* ... that is in use ... */
        if ( c.l[i].inuse )
        {
            /* write it to the file */
            for (k=0;k<7;k++) fputs ( xtsu_time_to_text ( &c.l[i].d[k].total ), fp );
            fputs ( c.l[i].name, fp );
            fputc ( '\n', fp );
        }
    }
        
    /* close the file */
    fclose(fp);
    
#ifdef VMS
    /*
    **  If we are on VMS, delete the old file.
    */
    if ( setup.purge )
    {
        char old[256];
        int status;
        
        strcpy ( old, filename );
        strcat ( old, ";-1" );
        status = remove ( old );
        if ( status != 0 )
        {
            char x[256];
            
            sprintf
            (
                x, 
                "Could not purge output file %s, status %d, errno %s", 
                old, 
                status,
                errno_text() 
            );
            xtsu_message ( "Purge Failed", x );
        }
    }
#else
    /*
    **  Rename the file thus getting rid of the existing one
    */
    rename ( tempfile, filename );
#endif

    /* update the crash recovery lock file */
    {
        datetime *lock;
        
        /* delete existing lock file */
        if ( xtsu_lock_file_delete() != 0 )
        {
            /*
            xtsu_message ( "Lock File Delete Failed", "Save was successful but the lock file could not be deleted.  This may not be important." );
            */
        }
        
        /* create new lock file */
        lock = xtsu_lock_file_create ();
    }

}

/*
**  Perform crash recovery
**  
**  (user has been prompted "do you want it done" and they have clicked on
**  either OK or Cancel...)
*/
static void xtso_charge_load_recover (w, change, reason)
Widget          w;
datetime        *change;
XmAnyCallbackStruct *reason;
{
    /* what did they click on? */
    switch ( reason->reason )
    {
        /* the OK button? */
        case XmCR_OK :
            
            /* perform the crash recovery */
            {
                int i, k;
    
                /* allocate next free entry */
                for (i=0;i<maxchg;i++) if ( ! c.l[i].inuse ) break;
                
                /* mark it in use */
                c.l[i].inuse = TRUE;
                
                /* set the name */
                /* ?? what if the name is already in use from a prior crash? */
                strcpy ( c.l[i].name, "crash recovery" );
                
                /* reset the times */
                for (k=0;k<7;k++) c.l[i].d[k].total = notime;
                
                /* add in the lost time to the current day */
                xtsu_time_add ( change, &c.l[i].d[c.day].total, &c.l[i].d[c.day].total );
    
                /* update the list */
                xtso_list_add ( i );
                xtso_list_totals ( );
            }
            break;
        
        /* or the Cancel button? */
        case XmCR_CANCEL :
            break;
    }
    
    /* destroy the message box widget */
    XtDestroyWidget ( w );
}

/*
**  Load Charge Table from File
*/
int xtso_charge_load(filename)
    char *filename;
{
    FILE *fp;
    int i, j, k;
    
    /* mark all entries free and clear the display list */
    for (i=0;i<maxchg;i++) c.l[i].inuse = FALSE ;
    xtso_list_clear ( );
    
    /* open the file */
    fp = fopen(filename,"r");
    
    /* if the file was not opened... */
    if ( fp == 0 )
    {
        /* create an empty charge list with the required default entry */
        i=0;
        strcpy ( c.l[i].name, xtsu_db_get_string ( "Xts.defaultEntryName", "default entry" ) );
        for (k=0;k<7;k++) c.l[i].d[k].total = notime;
        c.l[i].inuse = TRUE ;
        xtso_list_add ( i );
        if ( setup.totals ) xtso_list_totals_on ( );
        xtsu_lock_file_create ();
        return(FALSE);
    }
    
    /* file has been opened, read each record */
    for (i=0;i<maxchg;i++)
    {
        /* stop if end of file reached */
        if ( feof(fp) ) break ;
        
        /* read in the delta times for each day */
        for (k=0;k<7;k++)
        {
            char x[17];
            fgets ( x, 17, fp );
            if ( feof(fp) ) break ;
            memcpy ( &c.l[i].d[k].total, xtsu_time_from_text(x), sizeof(datetime));
        }
        
        /* check again for end of file */
        if ( feof(fp) ) break ;
        
        /* read in the charge entry description */
        fgets ( c.l[i].name, maxname, fp );
        j = strlen(c.l[i].name)-1;
        if ( c.l[i].name[j] == 10 ) c.l[i].name[j] = '\0';
        
        /* mark entry in use */
        c.l[i].inuse = TRUE ;

        /* add item at bottom */
        xtso_list_add ( i );
    }
    
    /* close the file */
    fclose(fp);

    /* display the totals */
    if ( setup.totals ) xtso_list_totals_on ( );
    
    /*
    **  Crash recovery
    **  
    **  If a lock file exists from a previous crash then it contains the date
    **  and time of the last save or load of an XTS charge file.  If the amount
    **  of uncharged time exceeds five minutes, then create an entry into which
    **  it can be placed.
    */
    {
        datetime *lock, now;
        static datetime change;  /* used by callback */
        
        lock = xtsu_lock_file_create ();
        xtsu_time_get ( &now );
        if ( memcmp ( lock, &now, sizeof ( datetime ) ) != 0 )
        {
            xtsu_time_subtract ( &now, lock, &change );
            if ( xtsu_time_hours ( &change ) > ( xtsu_db_get_int ( "Xts.crashRecoveryThresholdMinutes", 5 ) / 60.0 ) )
            {
                char text[256];
                sprintf ( text, "%-5.2f hours will be added as a crash recovery line", xtsu_time_hours ( &change ) );
                
                xtsu_confirm
                (
                    "Crash Recovery Confirmation",
                    text,
                    xtso_charge_load_recover,
                    &change
                );
            }
        }
    }
    
    return ( TRUE );
}

/*
**  Initialise the charging function
**  (save the current time)
*/
static void xtso_charge_init()
{
    xtsu_time_get ( &c.previous );
}

/*
**  Charge elapsed time to the current entry
*/
static void xtso_charge_to()
{
    int i;
    datetime now, change;
    
    /* select current charge list entry */
    i = c.now;

    /* get current time */
    xtsu_time_get ( &now );
    
    /* calculate elapsed time; now - previous */
    xtsu_time_subtract ( &now, &c.previous, &change );
    
    /* accumulate into entry */
    xtsu_time_add ( &change, &c.l[i].d[c.day].total, &c.l[i].d[c.day].total );
    
    /* update current item */
    xtso_list_update ( i );
    xtso_list_totals ( );

    /* save previous */
    c.previous = now;
}






/*
**  Window List
*/

/*
**  Scan the window list for a window,
**      Return either -1 for not-found or
**      the relative entry number in the window list.
*/
int xtso_window_list_scan ( key )
    Window key;
{
    int i;
    
    /* find window in list */
    for (i=0;i<maxwin;i++) if (window[i].id == key) return(i);
    return(-1);
}






/*
**  Return the current timer interval for focus monitoring
*/
static int xtso_focus_timer()
{
    int dummy;
    
    /*
    **  The actual time dependend on focus timer value and whether an
    **  association is pending.
    */
    if ( pending )
    {
        dummy = setup.focus_timer / 30;
    }
    else
    {
        dummy = setup.focus_timer;
    }
    
    /* force to not be below minimum for X server performance reasons */
    if ( dummy < minfocus ) dummy = minfocus;
    
    return ( dummy );
}


/*
**  Focus monitor timer work function
**  Manages detection of window focus changes
*/
static int xtso_timer_work ()
{
    Window now;                  /* current window in focus */
    static Window was;           /* previous window that was in focus */
    int dummy;                   /* place to put unused data */
    static int blinkstate = 0;   /* current blink state */
    
    /*
    **  Blink the current list item if an association is pending
    **  (the blinking continues until the association is resolved)
    */
    if ( pending )
    {
        if ( blinkstate == 0 )
        {
            XmListDeselectAllItems ( w_list );
            blinkstate++;
        }
        else
        {
            XmListSelectPos ( w_list, c.now+1, FALSE );
            blinkstate--;
            
            /* check for expiry of pending association */
            blinks--;
            if ( blinks == 0 ) pending = FALSE;
        }
    }
    
    /* reinsert ourselves into the timer queue */
    i_focus = XtAppAddTimeOut ( app_context, xtso_focus_timer(), xtso_timer_work, 0 );
    
    /* increment our counter */
    /* ?? remove counters someday, they are not being used */
    counts.focus.timer++;

    /* find the window that has the current input focus */
    XGetInputFocus (display,&now,&dummy);

    /* if nothing has changed, finish here */
    if ( was == now )
    {
        return(FALSE);
    }



    /*
    **  Focus has changed to a new window
    */
    {
        int i;
        char *win_name;
        
        /* don't worry about this window again; once is enough */
        was = now;
        
        /* if it's our own toplevel window, ignore the focus change */
        if ( now == wn_toplevel ) return (FALSE);
        
        /* if it is one of our other toplevel shell widgets, ignore it */
        for ( i=0; i<maxothers; i++ )
        {
            /* is end of list found? if so, stop looking */
            if ( wn_others[i] == 0 ) break;
            
            /* is the window one of ours? if so, ignore it */
            if ( wn_others[i] == now ) return (FALSE);
        }

        /*
        **  If the window hasn't got a name, explore upwards for a parent window
        **  that has a name and use that window instead.  This catches a queer
        **  problem with DECterm windows; the focus window varies according
        **  to whether the user clicked in the window, or if they used the
        **  window manager to pass focus to the window.
        **  
        **  In one of these situations, a child of the toplevel window for the
        **  DECterm has the input focus, and so XGetInputFocus() returns the
        **  child window.  We need to know the true identity of the window so
        **  that we can treat it the same no matter how the user got to it.
        */
        
        /*
        **  Keep looking for parents of the current window until we have found
        **  a window that has a name.  (Doesn't look for a parent unless the
        **  current window has no name.)
        */
        for (;;)
        {
            char *n;              /* the window name */
            Window root;          /* the root window */
            Window parent;        /* the parent of a window */
            Window *children;     /* the list of children of a window */
            unsigned int quiver;  /* the count of children; (a biblical term) */
                    
            /* initialise pointer to window name */
            n = 0;
            
            /* fetch the current window's name */
            XFetchName ( display, now, &n );
            
            /* is the name present? */
            if ( n )
            {
                /*
                **  ??
                **  
                **  Right here is where a filter of window names can be placed.
                **  The logic would be to compare the name fetched against
                **  a list of names, and to return(FALSE) if the window is to
                **  be ignored by XTS.
                */
                
                /* window has name, so stop here */
                XFree ( n );
                break;
            }
            else
            {
                /*
                **  ??
                **  
                **  Pointer focus policy buglet is here; no window may have
                **  input focus, and so XQueryTree will fail with invalid
                **  window ID.  Need some way to validate a window ID before
                **  trying to use it in any way.
                */
                
                /* window has no name, look up tree to parent */
                if ( ! XQueryTree ( display, now, &root, &parent, &children, &quiver ) )
                {
                    /* Query tree failed; default to the window we started with */
                    now = was;
                    break;
                }
                
                /*
                **  If a list of children was returned, ignore but free it.
                **  We are not interested in window children.
                */
                if (children) XFree((char *)children);
                
                /*
                **  If the parent window is the root window then we have
                **  drawn a blank.  We have reached the top of the window
                **  tree without finding a window with a name, so keep the
                **  window we had to start with.
                */
                if ( parent == root )
                {
                    now = was;
                    break;
                }
            
                /*
                **  Let's look at the parent now...
                */
                now = parent;
            }
        }

        /*
        **  At this point, the window is to be treated as genuine and observable
        **  by XTS; it has passed all the previous checks.
        */
        
        /* increment counter of focus changes */
        counts.focus.changes++;
        
        /* find window in list of windows */
        i = xtso_window_list_scan ( now );

        /* is window new? */
        if ( i == -1 )
        {
            /*
            **  Ignore this new window if there is no association pending,
            **  we are not interested in it and do not wish to record it.
            */
            if ( ! pending ) 
            {
                if ( setup.focus_revert != 0 )
                {
                    if ( setup.charging ) xtso_charge_to();
                    c.now = 0;
                    /* mark the new entry */
                    xtso_list_mark(c.now);
                }
                return ( FALSE );
            }
            
            /* find free slot in window list by searching for zero entry */
            i = xtso_window_list_scan ( 0 );
            if ( i != -1 )
            {
                /* create new entry in window list */
                counts.focus.new++;
                window[i].id = now;
                xtsu_time_get ( &window[i].seen );

                /* if window list is displayed, add a new entry to it */
                if ( w_window_list != 0 )
                    if ( XtIsManaged ( w_window_list ) )
                {
                    XmString s;
                    s = XmStringCreateSimple ( xtso_window_to_text ( i ) );
                    XmListAddItem ( w_window_list, s, 0 );
                    XmStringFree ( s );
                }
            }
        }
        else
        {
            /* seen this window before, note the time */
            xtsu_time_get ( &window[i].seen );
        }

        /*
        **  If an association is pending, then the current entry is to be
        **  connected to this window; do that now.
        */
        if ( pending )
        {
            /* connect this window to the current entry */
            window[i].charge = c.now;
            
            /* turn off pending association mode */
            pending = FALSE;
                
            /* if window list is displayed, make the relavent entry current */
            if ( w_window_list != 0 )
                if ( XtIsManaged ( w_window_list ) )
            {
                XmString s;
                s = XmStringCreateSimple ( xtso_window_to_text ( i ) );
                XmListReplaceItemsPos ( w_window_list, &s, 1, i+1 );
                XmListSelectPos ( w_window_list, i+1, TRUE );
                XmStringFree ( s );
            }
            
            /* ring the bell if requested */
            if ( setup.focus_bell ) XBell ( display, xtsu_db_get_int ( "Xts.focusBellRelativeVolume", -90 ) );
            
            /* make sure the entry is highlighted, then reset the blink state */
            XmListSelectPos ( w_list, c.now+1, FALSE );
            blinkstate = 0;
        }
        else
        {
            /*
            **  Just a normal focus change event; if the new window is for
            **  a different entry, accumulate the latest work into the old
            **  entry and begin charging to the new entry.
            */
            if ( window[i].charge >= 0 )
            {
                if ( setup.charging ) xtso_charge_to();
                c.now = window[i].charge;
                /* mark the new entry */
                xtso_list_mark(c.now);
            }

            /* if window list is displayed, select the current window entry */
            if ( w_window_list != 0 )
                if ( XtIsManaged ( w_window_list ) )
                    XmListSelectPos ( w_window_list, i+1, TRUE );
        }
        
        old = i;
    }
    return(FALSE);
}



/*
**  Main Window Charging Toggle Button Callback 
*/
void xtsu_charging(w, tag, reason)
Widget          w;
int             *tag;
XmToggleButtonCallbackStruct *reason;
{
    xtsu_copyright_off();
    /* the toggle button; is it being set or reset? */
    if ( reason->set )
    {
        /* being set, is charging off right now? */
        if ( ! setup.charging )
        {
            /* yes; so turn charging on */
            xtsu_new_day_check();
            setup.charging = TRUE;
            xtso_charge_init();
        }
    }
    else
    {
        /* being reset; is charging off right now? */
        if ( setup.charging )
        {
            /* finish charging that interval to the current entry */
            xtso_charge_to();
            
            /* mark it as current */
            xtso_list_mark(c.now);
            
            /* if the icon is to be changed, set it to default */
            if ( setup.icon ) xtso_set_icon ( xtsu_db_get_string ( "Xts.iconNameDefault", "XTS" ) );

            /* inform project file */
            if ( setup.project ) xtso_set_project ( xtsu_db_get_string ( "Xts.projectFileNotChargingString", "not charging to timesheet" ) );
            
            /* turn off charging */
            setup.charging = FALSE;
        }
    }
}

/*
**  Clear Week
**  
**  User has said yes, they are sure they know what they are doing, so
**  clear the entire week's accumulated charge time.
*/
void xtsu_clear_week_ok(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    int i,k;
    
    for (i=0;i<maxchg;i++)
    {
        if ( c.l[i].inuse )
        {
            for (k=0;k<7;k++) c.l[i].d[k].total = notime;
        }
    }
    xtso_list_update_all ( );
}

/*
**  Clear Day
**  
**  User has said yes, they are sure they know what they are doing, so
**  clear the day's accumulated charge time.
*/
void xtsu_clear_day_ok(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    int i;
    
    for (i=0;i<maxchg;i++)
    {
        if ( c.l[i].inuse )
        {
            c.l[i].d[c.day].total = notime;
        }
    }
    xtso_list_update_all ( );
}

/*
**  Copy a buffer to the clipboard
*/
static void xtsu_copy_to_clipboard ( char *buffer, int buflen, Time time )
{
    XmString s;
    int i, j;
    
    /* ? name of clipboard data */
    s = XmStringCreateSimple ( "xts" );
    
    /* start the copy */
    i = XmClipboardStartCopy
        (
          display,
          XtWindow(toplevel_widget),
          s,
          time,
          0,
          0,
          &j
        );
    XmStringFree ( s );
        
    i = XmClipboardCopy
        (
          display,
          XtWindow(toplevel_widget),
          j,
          "STRING",
          buffer,
          buflen,
          0,
          0
        );
        
    i = XmClipboardEndCopy
        (
          display,
          XtWindow(toplevel_widget),
          j
        );
}

void xtsu_export_cssts(w, tag, reason)
Widget          w;
int             *tag;
XmAnyCallbackStruct *reason;
{
    int i, k;
    char buffer[8192], line[256];
    int l, p, q;
    
    strcpy ( buffer, "" );
    p = 0;
    
    /* process each charge entry */
    for (i=0;i<maxchg;i++)
    {
        if ( c.l[i].inuse )
        {
            int j;
            datetime x;
            char v[7];
            
            strcpy  ( line, "" );
            strncat ( line, &c.l[i].name[0], 5 );
            strcat  ( line, "\t" );
            strncat ( line, &c.l[i].name[6], 3 );
            strcat  ( line, "\t" );
            
            for (j=0;j<7;j++)
            {
                sprintf ( v, "%-5.2f\t", xtsu_time_hours(&c.l[i].d[j].total) );
                strcat ( line, v );
            }
            strcat ( buffer, line );
        }
    }

    xtsu_copy_to_clipboard ( buffer, strlen ( buffer ), CurrentTime );
}

/*
**  Export in ITTS format
**  (old, probably nobody needs it now)
*/
void xtsu_export_itts_old(w, tag, reason)
Widget          w;
int             *tag;
XmAnyCallbackStruct *reason;
{
    int i, k;
    char buffer[4096], v[7];
    int l, p, q;
    
    p = 0;
    
    for (i=0;i<maxchg;i++)
    {
        if ( c.l[i].inuse )
        {
            /* key sequence for PF1 A */
            strncpy ( &buffer[p], "\x1bOPA", 4 );
            p = p + 4;
            q = 0;
            for (;;)
            {
                if ( c.l[i].name[q] == '\0' ) break;
                
                if ( c.l[i].name[q] == '\\' )
                {
                    buffer[p++] = '\n';
                    q++;
                }
                else
                {
                    buffer[p++] = c.l[i].name[q++];
                }
            }
            
            /* output the numbers one after another */
            for (k=0;k<7;k++)
            {
                buffer[p++] = '\t';
                sprintf ( v, "%-5.2f", xtsu_time_hours(&c.l[i].d[k].total) );
                l = strlen(v) - 1;
                strncpy ( &buffer[p], v, l );
                p = p + l;
            }

            buffer[p++] = '\n';
            q++;

        }
    }

    xtsu_copy_to_clipboard ( buffer, p, CurrentTime );
}

/*
**  Export to clipboard in DECplan time charge format
*/
void xtsu_export_decplan(w, tag, reason)
Widget          w;
int             *tag;
XmAnyCallbackStruct *reason;
{
    int i, k;
    char buffer[8192], line[256];
    int l, p, q;
    
    /*
    **  Place header into buffer
    */
    strcpy ( buffer, "\n\"DECplan Format Exported from X-Timesheets\"\n" );
    strcat ( buffer, "\"Derived from file " );
    strcat ( buffer, storage );
    strcat ( buffer, "\"\n" );
    
    /*
    **  Column headers, surrounded in quotes and separated with TAB characters
    */
    strcpy ( line, "" );
    strcat ( line, "\"Type\"\t\t" );
    strcat ( line, "\"Name\"\t" );
    strcat ( line, "\"Charged To\"\t" );
    strcat ( line, "\"Start\"\t" );
    strcat ( line, "\"Quantity Type\"\t" );
    strcat ( line, "\"Quantity Unit\"\t" );
    strcat ( line, "\"Quantity\"\t" );
    strcat ( line, "\"Posted\"\n" );
    
    strcat ( buffer, line );
    p = strlen(buffer);
    
    /* process each charge entry */
    for (i=0;i<maxchg;i++)
    {
        if ( c.l[i].inuse )
        {
            int j;
            datetime x;
            char v[7];
            
            strcpy ( line, "\"Time Charge\"\t0\t\"" );
            strcat ( line, c.l[i].name );
            strcat ( line, "-><RESOURCE>, Task Assignment\"\t" );
            strcat ( line, "<DATE>\t" );
            strcat ( line, "\"Effort\"\t" );
            strcat ( line, "\"Hours\"\t" );
            strcat ( buffer, line );
            
            strcpy ( line, "" );
            x = notime;
            for (j=0;j<7;j++) xtsu_time_add ( &c.l[i].d[j].total, &x, &x );
                
            sprintf ( v, "%-7.4f", xtsu_time_hours(&x) );
            strcat ( line, v );
            strcat ( line, "\t" );
            
            strcat ( line, "\"On\"\n" );
            strcat ( buffer, line );
        }
    }

    xtsu_copy_to_clipboard ( buffer, strlen ( buffer ), CurrentTime );
}

xtsu_new_day_check()
{
    int i;
    /*
    **  Check for new day
    */
    i = xtsu_work_day_number();
    if ( c.day != i )
    {
        /* set the new current day */
        c.day = i;
        
        /* is the work week mode enabled? */
        if ( setup.work_week_enable )
        {
            /* read the control file */
            xtsu_control_read ();
    
            /* is it a new week too? */
            if ( c.day < setup.day )
            {
                /* save the previous week's file */
                xtso_charge_save ( storage );
                
                /* move to next week */
                xtsu_control_increment();
    
                /* clear the data */
                xtsu_clear_week_ok ( 0, 0, 0 );
                
                /* set new file name */
                xtso_charge_name();
                
                {
                    char x[256];
                    sprintf ( x, "New fiscal week %02d of year %04d", setup.fiscal_week, setup.fiscal_year );
                    xtsu_message ( "New Fiscal Week", x );
                }
                
            } /* not new week */
        } /* work week mode disabled */
        
        /* write control file */
        setup.day = c.day;
        xtsu_control_write ();
                
        /* it's a new day anyway, so update the display */
        xtso_list_update_all();

    } /* not a new day */
}

/*
**  List update timer task
**  Update the current entry
**  Cope with new day occurrence
**  Cope with new week occurrence
*/
void xtsu_update_timer()
{
    /* resubmit ourselves for re-execution */
    i_update = XtAppAddTimeOut ( app_context, setup.update_timer, xtsu_update_timer, 0 );
    
    /* if charging is enabled, accumulate the interval to the current entry */
    if ( setup.charging )
    {
        xtso_charge_to();
        xtso_list_mark(c.now);
    }
    
    /* check for a new day */
    xtsu_new_day_check();
    
#ifdef VMS
    /* check for pause window */
    if ( setup.pause_action != 0 )
    {
        int is            = 0;   /* whether the pause window is on screen */
        static int paused = 0;   /* whether the pause was active on last call */
        static int entry  = 0;   /* charge entry that was current on pause */
        static int pass   = 0;   /* whether the pause threshold has passed */
        static datetime when;    /* time that pause started */
        
        is  = SmPauseWindow ( display,
            RootWindow ( display, DefaultScreen (display)) );
        
        if ( is != 0 )
        {
            if ( paused == 0 )
            {
                /* workstation has just been paused */
                paused = 1;
                pass = 0;
                xtsu_time_get ( &when );
            }
            
            /* detect passing of pause threshold */
            if ( pass == 0 )
            {
                datetime now;        /* current time */
                datetime elapsed;    /* time elapsed since pause start */
                
                xtsu_time_get ( &now );
                xtsu_time_subtract ( &now, &when, &elapsed );

                if ( elapsed >= setup.pause_thresh )
                {
                    /* pause window has been active for threshold time */
                    
                    /* adopt new position */
                    switch ( setup.pause_action )
                    {
                        case 1: /* stop charging */
                            
                            /* if we are charging */
                            if ( setup.charging )
                            {
                                Widget id;
                                /* take off pause time from current entry */
                                xtsu_time_subtract
                                (
                                    &c.l[c.now].d[c.day].total, 
                                    &elapsed, 
                                    &c.l[c.now].d[c.day].total
                                );
                                
                                HashLookup ( "w_tb_charging", &id) ;
                                XmToggleButtonGadgetSetState ( id, FALSE, TRUE ) ;
                            }
                            break;
                            
                        case 2: /* charge to top */
                            
                            /* save the current charge entry for unpause */
                            entry = c.now;
                            
                            /* if we are charging to non-top entry */
                            if ( setup.charging && c.now != 0 )
                            {
                                /* take off pause time from current entry */
                                xtsu_time_subtract
                                (
                                    &c.l[c.now].d[c.day].total, 
                                    &elapsed, 
                                    &c.l[c.now].d[c.day].total
                                );
                                
                                /* move to top entry */
                                c.now = 0;
                                xtso_list_mark ( c.now );
                            }
                            break;
                    }
                    pass = 1;
                }
            }
        }
        else
        {
            if ( paused != 0 )
            {
                /* workstation has just been unpaused */
                paused = 0;
                switch ( setup.pause_action )
                {
                    case 1: /* stop charging */
                    
                        /* resume charging */
                        if ( ! setup.charging )
                        {
                            Widget id;
                            
                            HashLookup ( "w_tb_charging", &id) ;
                            XmToggleButtonGadgetSetState ( id, TRUE, TRUE ) ;
                            xtso_list_mark ( c.now );
                        }
                        break;
                        
                    case 2: /* charge to top */
                    
                        /* revert to saved entry */
                        c.now = entry;
                        xtso_list_mark ( c.now );
                        break;
                }
            }
        }
    }
#endif /* VMS */
}


/*
**  Load a charge file.
**  
**  Called when the user has selected a file to load and has pressed the
**  Load pushbutton.  Does not change the current output file name.
*/
void xtsu_load(w, tag, reason)
Widget          w;
int             *tag;
XmFileSelectionBoxCallbackStruct *reason;
{
    char *e;
    int i;
    
    XmStringGetLtoR ( reason->value, XmSTRING_DEFAULT_CHARSET, &e );
    xtsu_busy ( w );
    XmUpdateDisplay ( w );
    i = xtso_charge_load ( e );
    if ( ! i )
    {
        char x[256];
        sprintf ( x, "File %s could not be opened.  Defaults assumed.", e );
        xtsu_message ( "Timesheet Cleared", x );
    }        
    xtsu_not_busy ( w );
    XtFree ( e );
}

/*
**  Save charge file
**  
**  Called when the user has selected a file name for saving and has pressed the
**  Save pushbutton.  Changes the current output file name.
*/
void xtsu_save(w, tag, reason)
Widget          w;
int             *tag;
XmFileSelectionBoxCallbackStruct *reason;
{
    char *e;
    XmStringGetLtoR ( reason->value, XmSTRING_DEFAULT_CHARSET, &e );
    xtsu_busy ( w );
    XmUpdateDisplay ( w );
    xtso_charge_save ( e );
    strcpy ( storage, e );
    xtsu_not_busy ( w );
    XtFree ( e );
}

/*
**  Perform a save to the current save file
*/
void xtsu_save_quick(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    xtsu_busy(w);
    XmUpdateDisplay(w);
    
    /* update the list prior to save */
    if ( setup.charging )
    {
        xtso_charge_to();
        xtso_list_mark ( c.now );
    }
    
    /* perform the save */
    xtso_charge_save (storage);
    xtsu_not_busy(w);
}

/*
**  Automatic Timesheet Save
**  
**  Save file update task - write out a new save file if charging is on
*/
xtsu_saver_timer()
{
    i_saver = XtAppAddTimeOut ( app_context, setup.saver_timer, xtsu_saver_timer, 0 );
    if ( setup.charging ) xtsu_save_quick(toplevel_widget, 0, 0);
}


/*
**  Apply changes to settings
**  
**  The "setup" structure contains the current settings, and the "will"
**  structure contains the changes to the settings.  This function compares
**  each setting and applies the change, if any.
*/
void xtsu_apply_settings()
{
    /* timesheet display */
    
    if ( setup.charging != will.charging )
    {
        Widget id;
        
        HashLookup ( "w_tb_charging", &id) ;
        XmToggleButtonGadgetSetState ( id, will.charging, TRUE ) ;
        
        /*
        ** (the action routine for the main window charging push button will
        ** update the setup.charging value, we don't need to do it here)
        */
    }
    
    /* list display mode 0=day, 1=week, 2=total */
    if ( setup.mode != will.mode )
    {
        setup.mode = will.mode;
        /* update entire list */
        xtso_list_update_all();
    }
    
    /* totals line enable */
    if ( setup.totals != will.totals )
    {
        if ( will.totals )
        {
            setup.totals = TRUE;
            xtso_list_totals_on ( );
        }
        else
        {
            setup.totals = FALSE;
            xtso_list_totals_off ( );
        }
    }
    
    /* list display format 0=hh:mm, 1=hh:mm:ss, 2=h.hh */
    if ( setup.format != will.format )
    {
        setup.format = will.format;
        /* update entire list */
        xtso_list_update_all();
    }

    /* list display update timer */
    if ( setup.update_timer != will.update_timer )
    {
        setup.update_timer = will.update_timer;
        XtRemoveTimeOut(i_update);
        i_update = XtAppAddTimeOut ( app_context, setup.update_timer, xtsu_update_timer, 0 );
    }

    /* pause window monitor */
    
    /* pause threshold time */
    setup.pause_action = will.pause_action;
    setup.pause_thresh = will.pause_thresh;

    /* window focus monitoring */
    
    /* focus monitoring enable */
    if ( setup.focus_enable != will.focus_enable )
    {
        Widget id;
        
        HashLookup ( "w_tb_focus", &id ) ;
        XmToggleButtonGadgetSetState ( id, will.focus_enable, TRUE ) ;
        
        /*
        ** (the action routine for the main window focus push button will
        ** update the setup.focus_enable value, we don't need to do it here)
        */
    }
    
    /* bell on new window focus */
    if ( setup.focus_bell != will.focus_bell )
    {
        setup.focus_bell = will.focus_bell;
    }
    
    /* focus timer interval */
    if ( setup.focus_timer != will.focus_timer )
    {
        setup.focus_timer = will.focus_timer;
        if ( setup.focus_enable )
        {
            XtRemoveTimeOut(i_focus);
            i_focus = XtAppAddTimeOut ( app_context, xtso_focus_timer(), xtso_timer_work, 0 );
        }
    }
    
    setup.focus_revert = will.focus_revert;
    setup.focus_delay  = will.focus_delay;
    
    /* automatic timesheet save */

    /* file save timer interval */
    if ( setup.saver_timer != will.saver_timer )
    {
        setup.saver_timer = will.saver_timer;
        
        /* is there one on the queue already?  if so, reschedule it */
        if ( setup.saver_enable )
        {
            XtRemoveTimeOut(i_saver);
            i_saver = XtAppAddTimeOut ( app_context, setup.saver_timer, xtsu_saver_timer, 0 );
        }
    }
    
    /* file save timer enable */
    if ( setup.saver_enable != will.saver_enable )
    {
        setup.saver_enable = will.saver_enable;
        if ( setup.saver_enable )
        {
            i_saver = XtAppAddTimeOut ( app_context, setup.saver_timer, xtsu_saver_timer, 0 );
        }
        else
        {
            XtRemoveTimeOut(i_saver);
        }
    }
    
    setup.icon = will.icon;
    setup.project = will.project;
    setup.purge = will.purge;
    
    /* work week mode */
    
    /* work week start */
    if ( setup.work_week_start != will.work_week_start )
    {
        int i;
        Widget id;
        
        setup.work_week_start = will.work_week_start;

        /* Determine current day */
        c.day = xtsu_work_day_number();
    
        /* if work week is currently enabled, and ... */
        if ( setup.work_week_enable )
        {
            /* is not being disabled ... */
            if ( will.work_week_enable )
            {
                /* write control file */
                setup.day = c.day;
                xtsu_control_write ();
            }
        }
        
        /* update the edit window */
        if ( HashLookup ( "w_fd_entry", &id ) )
            if ( XtIsManaged ( id ) ) xtsu_entry_labels();
        
        /* update the main window */
        if ( setup.mode == 0 ) xtso_list_update_all();

    }

    /* fiscal date */
    if (
        ( setup.fiscal_week != will.fiscal_week ) ||
        ( setup.fiscal_year != will.fiscal_year )
       )
    {
        setup.fiscal_week = will.fiscal_week;
        setup.fiscal_year = will.fiscal_year;
        
        /* if work week is currently enabled */
        if ( setup.work_week_enable )
        {
            /* and is not being disabled */
            if ( will.work_week_enable )
            {
                /* set file name */
                xtso_charge_name();
                
                /* write control file */
                setup.day = c.day;
                xtsu_control_write ();
            }
        }
    }
    
    /* work week enable */
    if ( setup.work_week_enable != will.work_week_enable )
    {
        int i;
        setup.work_week_enable = will.work_week_enable;

        if ( setup.work_week_enable )
        {
            /* read control file */
            xtsu_control_read ();

            /* set file name */
            xtso_charge_name();
        }
        else
        {
            /*
            **  Work week mode is being turned off.
            **  Ignore it.
            */
        }
    }
    
    
}

/*
**  Customise dialog box request
**  
**  Move the current customise settings to the customise window
**  (also used if the settings are changed while the window is active)
*/
void xtsu_customise(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    Widget      id;
    
    /* timesheet display */
    {
        char x[12];
        Widget a, b;
        
        HashLookup ( "w_om_mode", &a );
        sprintf ( x, "w_pb_mode_%d", setup.mode );
        HashLookup ( x, &b );
        xtsu_set_value ( a, XmNmenuHistory, b );
    }
    
    HashLookup ( "w_tb_totals", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.totals, TRUE ) ;

    {
        char x[12];
        Widget a, b;
        
        HashLookup ( "w_om_format", &a );
        sprintf ( x, "w_pb_format_%d", setup.format );
        HashLookup ( x, &b );
        xtsu_set_value ( a, XmNmenuHistory, b );
    }
    
    HashLookup ( "w_sl_update_timer", &id) ;
    XmScaleSetValue ( id, setup.update_timer / 60000 );

    /* pause window */
    
    {
        char x[12];
        Widget a, b;
        
        HashLookup ( "w_om_pause_action", &a );
        sprintf ( x, "w_pb_pause_action_%d", setup.pause_action );
        HashLookup ( x, &b );
        xtsu_set_value ( a, XmNmenuHistory, b );
    }
    
    HashLookup ( "w_sl_pause_thresh", &id) ;
    XmScaleSetValue ( id, setup.pause_thresh / 60 );

    /* window focus monitoring */
    
    HashLookup ( "w_tb_focus_enable", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.focus_enable, TRUE ) ;

    HashLookup ( "w_tb_bell", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.focus_bell, TRUE ) ;

    HashLookup ( "w_sl_timer", &id) ;
    XmScaleSetValue ( id, setup.focus_timer / 100 );

    HashLookup ( "w_tb_focus_revert", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.focus_revert, TRUE );

    HashLookup ( "w_sl_focus_delay", &id) ;
    XmScaleSetValue ( id, setup.focus_delay / 100 );

    /* automatic timesheet save */
    HashLookup ( "w_tb_saver_enable", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.saver_enable, TRUE ) ;

    HashLookup ( "w_sl_saver_timer", &id) ;
    XmScaleSetValue ( id, setup.saver_timer / 60000 );

    HashLookup ( "w_tb_icon", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.icon, TRUE ) ;

    HashLookup ( "w_tb_project", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.project, TRUE ) ;

    HashLookup ( "w_tb_work_week_enable", &id) ;
    XmToggleButtonGadgetSetState ( id, setup.work_week_enable, TRUE ) ;

    /* set current state of work week menu */
    {
        char x[12];
        Widget a, b;
        
        HashLookup ( "w_om_work_week_start", &a );
        sprintf ( x, "w_pb_work_%d", setup.work_week_start );
        HashLookup ( x, &b );
        xtsu_set_value ( a, XmNmenuHistory, b );
    }
    
    /* set current state of fiscal year field */
    {
        char x[5];
        
        sprintf ( x, "%04d", setup.fiscal_year );
        HashLookup ( "w_tf_fiscal_year", &id );
        XmTextFieldSetString ( id, x );
    }

    /* set current state of fiscal week field */
    {
        char x[5];
        
        sprintf ( x, "%02d", setup.fiscal_week );
        HashLookup ( "w_tf_fiscal_week", &id );
        XmTextFieldSetString ( id, x );
    }

}

/*
**  Fetch the current settings from the customise window
*/
void xtsu_customise_apply(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    Widget      id;

    /* timesheet display */
    
    /*
    ** (this one doesn't have a customise button, so we have to just simulate
    ** grabbing the value from the window...)
    */
    will.charging = setup.charging;
    
    /* display mode */
    {
        char *x;
        Widget a, b;
        
        HashLookup ( "w_om_mode", &a ) ;
        xtsu_get_value ( a, XmNmenuHistory, &b );
        x = XtName ( b );
        sscanf ( x, "w_pb_mode_%d", &will.mode );
        XtFree ( x );
    }
    
    HashLookup ( "w_tb_totals", &id) ;
    will.totals = XmToggleButtonGadgetGetState ( id ) ;
    
    {
        char *x;
        Widget a, b;
        
        HashLookup ( "w_om_format", &a ) ;
        xtsu_get_value ( a, XmNmenuHistory, &b );
        x = XtName ( b );
        sscanf ( x, "w_pb_format_%d", &will.format );
        XtFree ( x );
    }
    
    HashLookup ( "w_sl_update_timer", &id) ;
    XmScaleGetValue ( id, &will.update_timer );
    will.update_timer = will.update_timer * 60000 ;

    /* pause window */
    
    {
        char *x;
        Widget a, b;
        
        HashLookup ( "w_om_pause_action", &a ) ;
        xtsu_get_value ( a, XmNmenuHistory, &b );
        x = XtName ( b );
        sscanf ( x, "w_pb_pause_action_%d", &will.pause_action );
        XtFree ( x );
    }
    
    HashLookup ( "w_sl_pause_thresh", &id) ;
    XmScaleGetValue ( id, &will.pause_thresh );
    will.pause_thresh = will.pause_thresh * 60 ;
    
    /* window focus monitoring */

    HashLookup ( "w_tb_focus_enable", &id) ;
    will.focus_enable = XmToggleButtonGadgetGetState ( id ) ;

    HashLookup ( "w_tb_bell", &id) ;
    will.focus_bell = XmToggleButtonGadgetGetState ( id ) ;

    HashLookup ( "w_sl_timer", &id) ;
    XmScaleGetValue ( id, &will.focus_timer );
    will.focus_timer = will.focus_timer * 100 ;
    
    HashLookup ( "w_tb_focus_revert", &id) ;
    will.focus_revert = XmToggleButtonGadgetGetState ( id ) ;

    HashLookup ( "w_sl_focus_delay", &id) ;
    XmScaleGetValue ( id, &will.focus_delay );
    will.focus_delay = will.focus_delay * 100 ;
    
    /* automatic timesheet save */
    HashLookup ( "w_tb_saver_enable", &id) ;
    will.saver_enable = XmToggleButtonGadgetGetState ( id ) ;

    HashLookup ( "w_sl_saver_timer", &id) ;
    XmScaleGetValue ( id, &will.saver_timer );
    will.saver_timer = will.saver_timer * 60000 ;

    HashLookup ( "w_tb_icon", &id) ;
    will.icon = XmToggleButtonGadgetGetState ( id ) ;

    HashLookup ( "w_tb_project", &id) ;
    will.project = XmToggleButtonGadgetGetState ( id ) ;

    /* get work week enable flag */
    HashLookup ( "w_tb_work_week_enable", &id) ;
    will.work_week_enable = XmToggleButtonGadgetGetState ( id ) ;

    /* get current state of work week menu */
    {
        char *x;
        Widget a, b;
        
        HashLookup ( "w_om_work_week_start", &a ) ;
        xtsu_get_value ( a, XmNmenuHistory, &b );
        x = XtName ( b );
        sscanf ( x, "w_pb_work_%d", &will.work_week_start );
        XtFree ( x );
    }
    
    /* get current state of fiscal year field */
    {
        char *a;
        
        HashLookup ( "w_tf_fiscal_year", &id ) ;
        a = XmTextFieldGetString ( id );
        sscanf ( a, "%d", &will.fiscal_year );
        XtFree ( a );
    }

    /* get current state of fiscal week field */
    {
        char *a;
        
        HashLookup ( "w_tf_fiscal_week", &id ) ;
        a = XmTextFieldGetString ( id );
        sscanf ( a, "%d", &will.fiscal_week );
        XtFree ( a );
    }

    xtsu_apply_settings();
}






/*
**  Window List Display
*/

/*
**  Delete an entry from the window list
*/
void xtsu_window_list_delete()
{
    if ( deleting != -1 )
    if ( window[deleting].id != 0 )
    {
        int i;
        
        for (i=deleting;i<(maxwin-1);i++) window[i] = window[i+1];
        XmListDeletePos ( w_window_list, deleting+1 );
    }
}

/*
**  Selet a window in the window list
*/
void xtsu_window_list_select(w, tag, reason)
Widget          w;
int             *tag;
XmListCallbackStruct *reason;
{
    deleting = reason->item_position - 1;
}

/*
**  Remember the widget id of the window list
*/
void xtsu_window_list_create(w, tag, reason)
Widget          w;
int             *tag;
unsigned long *reason;
{
    xts_register_widget(w,tag,reason);
    w_window_list = w;
}

/*
**  Move internal window list to displayed window list
*/
void xtsu_window_list()
{
    int i;
    
    XmListDeleteAllItems ( w_window_list );
    
    for (i=0;i<maxwin;i++)
    {
        if (window[i].id != 0)
        {
            XmString s;
            s = XmStringCreateSimple ( xtso_window_to_text ( i ) );
            XmListAddItem ( w_window_list, s, 0 );
            XmStringFree ( s );
        }
    }

    deleting = -1;
}





/*
**  Entry Editor
*/

/*
**  Fix the label strings of the vertical table
*/
static xtsu_entry_labels()
{
    static int was = -1;   /* last known work week start day number */
    
    if ( was != setup.work_week_start )
    {
        int i;
        
        was = setup.work_week_start;
        
        for ( i=0; i<7; i++ )
        {
            char x[12];
            Widget a, b;
            XmString s;
            
            /* ensure the customise window has been read from the UID file */
            VUIT_Read ( "w_fd_customise" );
            
            /* get label of work week menu widget for specified day */
            sprintf ( x, "w_pb_work_%d", ( was + i ) % 7 );
            HashLookup ( x, &a );
            xtsu_get_value ( a, XmNlabelString, &s );
            
            /* set label of headings label widget */
            sprintf ( x, "w_l_day_%d", i );
            HashLookup ( x, &a );
            xtsu_set_value ( a, XmNlabelString, s );
    
            /* finished with the string value */
            XmStringFree ( s );
        }
    }
}

/*
**  Entry editor request; from entry menu or by double click on charge list,
**  or during entry create...
**  
**  Moves entry data into entry editor dialog box
*/
void xtsu_entry()
{
    Widget      id;
    int i;
    
    /* cancel pending focus association */
    pending = FALSE;
    
    /* move entry being edited to the entry window */
    HashLookup ( "w_tf_name", &id ) ;
    XmTextFieldSetString ( id, c.l[editing].name );
    
    /* position cursor to the entry name window */
    XmProcessTraversal ( id, XmTRAVERSE_CURRENT );
    
    /* do the XmProcessTraversal() twice due to a Motif bug... */
    XmProcessTraversal ( id, XmTRAVERSE_CURRENT );

    /* put the current value into the current value field */
    HashLookup ( "w_tf_value", &id ) ;
    XmTextFieldSetString ( id, xtso_charge_value_to_text ( &c.l[editing].d[c.day].total ) );

    /* update labels */
    xtsu_entry_labels();

    /* move the remaining values into the fields */
    for (i=0;i<7;i++)
    {
        char x[20];
        
        sprintf ( x, "w_tf_value_%d", i );
        HashLookup ( x, &id ) ;
        if ( i == c.day )
        {
            XmTextFieldSetString ( id, xtsu_db_get_string ( "Xts.entryEditorInactiveString", "( see above )" ) );
            XtSetSensitive ( id, FALSE );
        }
        else
        {
            XmTextFieldSetString ( id, xtso_charge_value_to_text ( &c.l[editing].d[i].total ) );
            XtSetSensitive ( id, TRUE );
        }
    }
}

/*
**  Entry Editor Apply pushbutton pressed
*/
void xtsu_entry_apply(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    Widget      id;
    char *a;
    int i;
    
    /* fetch entry name */
    HashLookup ( "w_tf_name", &id ) ;
    a = XmTextFieldGetString ( id );
    
    /* if name omitted, don't do anythng */
    if ( ( strlen ( a ) == 0 ) && ( ! c.l[editing].inuse ) )
    {
        XtFree ( a );
        return;
    }
    
    /* set entry name */
    strncpy ( c.l[editing].name, a, maxname );
    XtFree ( a );
    
    /* set entry value */
    HashLookup ( "w_tf_value", &id ) ;
    a = XmTextFieldGetString ( id );
    memcpy ( &c.l[editing].d[c.day].total, xtso_text_to_charge_value(a), sizeof(datetime));
    XtFree ( a );
    
    /* pull off other day's values */
    for (i=0;i<7;i++)
    {
        char x[20];
            
        sprintf ( x, "w_tf_value_%d", i );
        HashLookup ( x, &id ) ;
        if ( XtIsSensitive ( id ) )
        {
            a = XmTextFieldGetString ( id );
            memcpy ( &c.l[editing].d[i].total, xtso_text_to_charge_value(a), sizeof(datetime));
            XtFree ( a );
        }
    }

    /* update the screen */
    if ( ! c.l[editing].inuse )
    {
        /* new entry being created */
        c.l[editing].inuse = TRUE;
        xtso_list_add ( editing );
    }
    else
    {
        xtso_list_update ( editing );
    }
    
    /* update totals if required */
    xtso_list_totals ( );
}

/*
**  Entry Create request from entry menu
*/
void xtsu_entry_create(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    Widget      id;
    int i, k;
        
    /* allocate next free entry */
    for (i=0;i<maxchg;i++) if ( ! c.l[i].inuse ) break;
    editing = i;
    
    /* clear entry */
    c.l[editing].name[0] = '\0';
    for (k=0;k<7;k++) c.l[editing].d[k].total = notime;

    /* put entry in window */
    xtsu_entry();
}

/*
**  Entry delete confirmation; do the delete
*/
void xtsu_entry_delete_ok(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    int n;
    
    /* get entry number to be deleted */
    n = c.now;
    
    if ( c.l[n].inuse )
    {
        int i;
        
        /*
        **  Delete the charge list entry by shifting the entries up.
        */
        for (i=n;i<(maxchg-1);i++)
        {
            c.l[i] = c.l[i+1];
        }
        
        /*
        **  Adjust the window list; decrement the charge id for every window
        **  entry that refers to a charge line on or beyond the one we have
        **  just deleted.
        */
        for (i=0;i<maxwin;i++) if ( window[i].id >= n ) window[i].id--;
        
        /*
        **  Delete the entry from the screen
        */
        XmListDeletePos ( w_list, n+1 );
        
        /*
        **  Update the totals line
        */
        xtso_list_totals ( );
        
    }

    /* resort to top entry */
    c.now = 0;
    xtso_list_mark ( c.now );
    
    /* reset pending association */
    pending = FALSE;
}

/*
**  Entry edit from entry menu
*/
void xtsu_entry_edit(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    Widget      id;
    
    editing = c.now;
    
    xtsu_entry();
}

/*
**  Called during creation of each major dialog box in the application,
**  causes the window ID of the dialog box to be saved in a list so that
**  XTS can ignore it's own windows.
*/
void xtsu_ignore_this_window(w, tag, reason)
Widget          w;
int             *tag;
unsigned long *reason;
{
    static int next = 0;
    
    if ( next < maxothers ) wn_others[next++] = XtWindow(XtParent(w));
}

void xtsu_quit(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    if ( xtsu_lock_file_delete() != 0 )
    {
        printf ( "xtsu_quit: failed to delete lock file\n" );
    }
}

void xtsu_exit(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    xtsu_busy(w);
    XmUpdateDisplay(w);
    xtso_charge_save (storage);
    if ( xtsu_lock_file_delete() != 0 )
    {
        printf ( "xtsu_exit: failed to delete lock file\n" );
    }
    xtsu_not_busy(w);
}

/*
**  Main Window Focus Monitor Pushbutton Callback
*/
void xtsu_focus(w, tag, reason)
Widget          w;
int             *tag;
XmToggleButtonCallbackStruct *reason;
{
    xtsu_copyright_off();
    pending = FALSE;
    if ( reason->set )
    {
        if ( ! setup.focus_enable )
        {
            setup.focus_enable = TRUE;
            i_focus = XtAppAddTimeOut ( app_context, xtso_focus_timer(), xtso_timer_work, 0 );
        }
    }
    else
    {
        if ( setup.focus_enable )
        {
            setup.focus_enable = FALSE;
            XtRemoveTimeOut(i_focus);
        }
    }
}


/*
**  Charge List Display
*/

/*
**  Create charge list - save widget id for further use
*/
void xtsu_list_create(w, tag, reason)
Widget          w;
int             *tag;
unsigned long *reason;
{
    w_list = w;
}

/*
**  Entry selected callback
**  
**  Called when the user selects an entry in the charge list in the main window.
*/
void xtsu_list(w, tag, reason)
Widget          w;
int             *tag;
XmListCallbackStruct *reason;
{
    int n;
    
    xtsu_copyright_off();
    n = reason->item_position - 1;
    
    if ( setup.charging ) xtso_charge_to();
    if ( c.l[n].inuse ) c.now = n;
    xtso_list_mark ( c.now );
    
    /* if focus monitoring enabled, start a pending association */
    if ( setup.focus_enable )
    {
        pending = TRUE;
        blinks = xtsu_db_get_int ( "Xts.maxBlinksWhilePending", 30 );

        XtRemoveTimeOut(i_focus);
        i_focus = XtAppAddTimeOut ( app_context, setup.focus_delay, xtso_timer_work, 0 );
    }
}

/* set (or create) database to defaults */
void xtsu_read_defaults()
{
    will.charging         = xtsu_db_get_int ( "Xts.charging",          0 );
    will.work_week_enable = xtsu_db_get_int ( "Xts.work_week_enable",  1 );
    will.work_week_start  = xtsu_db_get_int ( "Xts.work_week_start",   1 );
    will.icon             = xtsu_db_get_int ( "Xts.icon",              1 );
    will.project          = xtsu_db_get_int ( "Xts.project",           0 );
    will.purge            = xtsu_db_get_int ( "Xts.purge",             1 );
    will.mode             = xtsu_db_get_int ( "Xts.mode",              0 );
    will.totals           = xtsu_db_get_int ( "Xts.totals",            0 );
    will.format           = xtsu_db_get_int ( "Xts.format",            0 );
    will.update_timer     = xtsu_db_get_int ( "Xts.update_timer",      60000    /* 60s   */ );
    will.pause_action     = xtsu_db_get_int ( "Xts.pause_action",      0 );
    will.pause_thresh     = xtsu_db_get_int ( "Xts.pause_thresh",      600      /* 10m   */ );
    will.focus_enable     = xtsu_db_get_int ( "Xts.focus_enable",      0 );
    will.focus_bell       = xtsu_db_get_int ( "Xts.focus_bell",        1 );
    will.focus_timer      = xtsu_db_get_int ( "Xts.focus_timer",       2300     /* 2.3s  */ );
    will.focus_revert     = xtsu_db_get_int ( "Xts.focus_revert",      0 );
    will.focus_delay      = xtsu_db_get_int ( "Xts.focus_delay",       0 );
    will.saver_enable     = xtsu_db_get_int ( "Xts.saver_enable",      1 );
    will.saver_timer      = xtsu_db_get_int ( "Xts.saver_timer",       900000   /* 15min */ );
}

void xtsu_use_defaults()
{
    Widget id;
    
    xtsu_read_defaults();
    xtsu_apply_settings();
    
    /* update the customise dialog box if it is on screen */
    if ( HashLookup ( "w_fd_customise", &id ) )
        if ( XtIsManaged ( id ) ) xtsu_customise(id,0,0);
}

void xtsu_save_settings(w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    Arg a[10];
    int i, vw, vh, vx, vy;
    XrmValue v;
    char t[256];
    
    xtsu_busy(w);
    XmUpdateDisplay ( w );
    
    /* XtGetValues seems to work better if arguments are initialised */
    vw = 0;
    vh = 0;
    vx = 0;
    vy = 0;
    
    i = 0;
    XtSetArg ( a[i], XmNwidth, &vw ) ; i++ ;
    XtSetArg ( a[i], XmNheight, &vh ) ; i++ ;
    XtSetArg ( a[i], XmNx, &vx ) ; i++ ;
    XtSetArg ( a[i], XmNy, &vy ) ; i++ ;
    XtGetValues ( toplevel_widget, a, i );
    
    xtsu_db_put_int ( "Xts.width",  vw );
    xtsu_db_put_int ( "Xts.height", vh );
    xtsu_db_put_int ( "Xts.x",      vx );
    xtsu_db_put_int ( "Xts.y",      vy );
    
    xtsu_db_put_int ( "Xts.charging", setup.charging );
    xtsu_db_put_int ( "Xts.mode", setup.mode );
    xtsu_db_put_int ( "Xts.totals", setup.totals );
    xtsu_db_put_int ( "Xts.format", setup.format );
    xtsu_db_put_int ( "Xts.update_timer", setup.update_timer );
    xtsu_db_put_int ( "Xts.pause_action", setup.pause_action );
    xtsu_db_put_int ( "Xts.pause_thresh", setup.pause_thresh );
    xtsu_db_put_int ( "Xts.focus_enable", setup.focus_enable );
    xtsu_db_put_int ( "Xts.focus_bell", setup.focus_bell );
    xtsu_db_put_int ( "Xts.focus_timer", setup.focus_timer );
    xtsu_db_put_int ( "Xts.focus_revert", setup.focus_revert );
    xtsu_db_put_int ( "Xts.focus_delay", setup.focus_delay );
    xtsu_db_put_int ( "Xts.saver_enable", setup.saver_enable );
    xtsu_db_put_int ( "Xts.saver_timer", setup.saver_timer );
    xtsu_db_put_int ( "Xts.icon", setup.icon );
    xtsu_db_put_int ( "Xts.project", setup.project );
    xtsu_db_put_int ( "Xts.work_week_enable", setup.work_week_enable );
    xtsu_db_put_int ( "Xts.work_week_start", setup.work_week_start );
    
    xtsu_db_save_current ();
    xtsu_not_busy(w);
}

void xtsu_restore_settings()
{
    Widget id;

    will.charging         = xtsu_db_get_int ( "Xts.charging",         0 );
    will.mode             = xtsu_db_get_int ( "Xts.mode",             0 );
    will.totals           = xtsu_db_get_int ( "Xts.totals",           0 );
    will.format           = xtsu_db_get_int ( "Xts.format",           0 );
    will.update_timer     = xtsu_db_get_int ( "Xts.update_timer",     0 );
    will.pause_action     = xtsu_db_get_int ( "Xts.pause_action",     0 );
    will.pause_thresh     = xtsu_db_get_int ( "Xts.pause_thresh",     0 );
    will.focus_enable     = xtsu_db_get_int ( "Xts.focus_enable",     0 );
    will.focus_bell       = xtsu_db_get_int ( "Xts.focus_bell",       0 );
    will.focus_timer      = xtsu_db_get_int ( "Xts.focus_timer",      0 );
    will.focus_revert     = xtsu_db_get_int ( "Xts.focus_revert",     0 );
    will.focus_delay      = xtsu_db_get_int ( "Xts.focus_delay",      0 );
    will.purge            = xtsu_db_get_int ( "Xts.purge",            0 );
    will.saver_enable     = xtsu_db_get_int ( "Xts.saver_enable",     0 );
    will.saver_timer      = xtsu_db_get_int ( "Xts.saver_timer",      0 );
    will.icon             = xtsu_db_get_int ( "Xts.icon",             0 );
    will.project          = xtsu_db_get_int ( "Xts.project",          0 );
    will.work_week_enable = xtsu_db_get_int ( "Xts.work_week_enable", 0 );
    will.work_week_start  = xtsu_db_get_int ( "Xts.work_week_start",  0 );

    /*
    **  Validate some of the settings
    */
#define xtsu_restore_settings_range(name,min,max) \
    if ( will.name < min || will.name > max ) will.name = min;
        
    xtsu_restore_settings_range(mode,         0,   3);
    xtsu_restore_settings_range(format,       0,   3);
    xtsu_restore_settings_range(focus_delay,  0, 100);
    xtsu_restore_settings_range(pause_action, 0,   2);
    xtsu_restore_settings_range(pause_thresh, 0, 240);
    
    xtsu_apply_settings();
    
    /* update the customise dialog box if it is on screen */
    if ( HashLookup ( "w_fd_customise", &id ) )
        if ( XtIsManaged ( id ) ) xtsu_customise(id,0,0);
}

/*
**  Registration of a message box widget for which only the OK button is present
**  This function removes the other buttons.
*/
xtsu_bare_message_box (w, tag, reason)
Widget          w;
int             *tag;
unsigned long   *reason;
{
    /*
    **  Get the help and cancel button widget id and unmanage them
    */
    XtUnmanageChild ( XmMessageBoxGetChild ( w, XmDIALOG_HELP_BUTTON ) );
    XtUnmanageChild ( XmMessageBoxGetChild ( w, XmDIALOG_CANCEL_BUTTON ) );
}


static int xtsu_error_handler ( dpy, rep )
    Display *dpy;
    XErrorEvent *rep;
{
    if (rep->error_code == BadWindow) { return 0; }
    (*xtsu_old_error_handler) ( dpy, rep );
}

static int xtsu_io_error_handler ( dpy )
    Display *dpy;
{
    if ( setup.charging )
    {
        xtso_charge_to();
        xtso_charge_save ( "crash.xts" );
        printf ( "xts: Fatal Xlib I/O error - timesheet saved in crash.xts\n" );
    }
    (*xtsu_old_io_error_handler) ( dpy );
}

/*
**  Main application initialisation; called immediately prior to the
**  Motif main loop.
*/
void xtsu_start ( file, clear )
    char *file;
    char *clear;
{
    /*
    **  Set error handlers
    */
    xtsu_old_error_handler    = XSetErrorHandler   ( xtsu_error_handler );
    xtsu_old_io_error_handler = XSetIOErrorHandler ( xtsu_io_error_handler );
    
    /*
    **  Set application icon
    */
    {
        Pixmap          pixmap;
        Dimension       width, height;      
    
        if ( MrmFetchBitmapLiteral ( 
                    s_MrmHierarchy,
                    "l_icon",       /* name of icon literal */
                    XtScreen(toplevel_widget),      /* screen pointer       */
                    display,
                    &pixmap,
                    &width,                 
                    &height ) ==  MrmSUCCESS )
        {
            /* set the icon pixmap */
            xtsu_set_value ( toplevel_widget, XmNiconPixmap, pixmap );
        }
    }

    /*
    **  Pre-read major dialog boxes to speed future activity
    */
    XtAppAddTimeOut ( app_context, 1500, VUIT_Read, "w_fd_customise" );
    XtAppAddTimeOut ( app_context, 2000, VUIT_Read, "w_fd_entry" );
    
    /*
    **  Initialise program default behaviour
    **  
    **  These defaults should be representative of the current state of things
    **  when the program begins, *not* the user oriented defaults that are
    **  described in the xtsu_read_defaults() function.  This is so that the
    **  xtsu_apply_settings() function will actually apply the defaults in order
    **  to get things running how they should.
    */
    
    /* timesheet display */
    setup.charging = 0;
    setup.mode = 0;              /* list display mode 0=day, 1=week, 2=total */
    setup.totals = 0;            /* totals line enable */
    setup.format = 0;            /* list display format 0=hh:mm, 1=hh:mm:ss, 2=h.hh */
    setup.update_timer = -1;     /* list display update timer */
    
    /* pause window */
    setup.pause_action = 0;
    setup.pause_thresh = 0;
    
    /* window focus monitoring */
    setup.focus_enable = 0;      /* focus monitoring enable */
    setup.focus_bell = 1;        /* bell on new window focus */
    setup.focus_timer = -1;      /* focus timer interval */
    setup.focus_revert = 0;      /* charge to default on new window? */
    setup.focus_delay  = 0;      /* delay on new association */
    
    /* automatic timesheet save */
    setup.saver_enable = 0;      /* file save timer enable */
    setup.saver_timer = -1;      /* file save timer interval */
    
    /* whether to use icon */
    setup.icon = 0;
    setup.project = 0;
    setup.purge = 0;
    
    setup.work_week_enable = 0;
    setup.work_week_start = -1;   /* not set */
    
    setup.fiscal_year = -1;
    setup.fiscal_week = -1;
    
    /* Determine current day */
    c.day = xtsu_work_day_number();
    
    /* Initialise the totals line title */
    strcpy ( c.t.name, xtsu_db_get_string ( "Xts.totalsLabelString", "Totals" ) );
    
    /*
    **  Initialise pointer to the window list widget
    */
    w_window_list = 0;
    
    /*
    **  Initialise the null time value.  This is done by getting the OS to
    **  give us the number for zero delta time, which is 0.1 microseconds for
    **  OpenVMS and 0 seconds for UNIX.
    */
    xtsu_time_get ( &notime );
    xtsu_time_subtract ( &notime, &notime, &notime );
    
    /*
    **  Obtain the window id of the toplevel widget so that we can ignore
    **  focus changes to that window
    */
    wn_toplevel = XtWindow(toplevel_widget);
    
    /* Create the wait cursor */
    cursor = DXmCreateCursor ( toplevel_widget, decw$c_wait_cursor );

    /* Restore saved settings (or default) */
    xtsu_restore_settings ( );
    
    if ( setup.work_week_enable )
    {
        FILE *cf;
            
        /* read the control file */
        xtsu_control_read ();

        /* set file name */
        xtso_charge_name();
        
        /* work out whether this is a new week */
        if ( c.day < setup.day )
        {
            /* it is a new week */
            
            /* read the old charge file */
            xtso_charge_load ( storage );
            
            /* move to next week */
            xtsu_control_increment();

            /* clear the data */
            xtsu_clear_week_ok ( 0, 0, 0 );
            
            /* set new file name */
            xtso_charge_name();
            
            /* save new file */
            xtso_charge_save ( storage );

            {
                char x[256];
                sprintf ( x, "New fiscal week %02d of year %04d", setup.fiscal_week, setup.fiscal_year );
                xtsu_message ( "New Fiscal Week", x );
            }
            
        }
        else
        {
            xtso_charge_load ( storage );
        }

        setup.day = c.day;
        xtsu_control_write ();
    }
    else
    {
        if ( file == 0 )
        {
            strcpy ( storage, xtsu_db_get_string ( "Xts.defaultFileName", "XTS_SAVE" ) );
        }
        else
        {
            strcpy ( storage, file );
        }
        xtso_charge_load ( storage );
    }
    
    /* if command line specified clear week, then do so now */
    if ( clear != 0 ) if ( ! strcmp ( clear, "clear" ) )
    {
        xtsu_clear_week_ok ( 0, 0, 0 );
    }
}
