#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifdef VMS
#include <types.h>
#else
#include <sys/types.h>
#endif
#include <time.h>
#include <Xm/MessageB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Mrm/MrmAppl.h>
#include <X11/X.h>
#include <DXm/DECspecific.h>
#ifdef VMS
#include <DECw$Cursor.h>
#else
#include <X11/decwcursor.h>
#endif
#include "XTS.h"

String title = NULL;

void xtsu_copyright_off();

static void xtsu_copyright_event_remove ( char *name )
{
    Widget id;
    if ( HashLookup ( name, &id ) )
    {
        XtRemoveEventHandler ( id, ButtonPressMask, FALSE, 
            (XtEventHandler) xtsu_copyright_off, 0 );
    }
    else
    {
        printf ( "xtsu_copyright_event_remove: unregistered widget %s\n", name );
    }
}

static void xtsu_copyright_event_add ( char *name )
{
    Widget id;
    if ( HashLookup ( name, &id ) )
    {
        XtAddEventHandler ( id, ButtonPressMask, FALSE, 
            (XtEventHandler) xtsu_copyright_off, 0 );
    }
    else
    {
        printf ( "xtsu_copyright_event_add: unregistered widget %s\n", name );
    }
}

void xtsu_copyright_off ( )
{
    Arg arglist[2];
    
    if ( title == NULL ) return;
    
    xtsu_copyright_event_remove ( "w_mb_main" );
/*    xtsu_copyright_event_remove ( "w_lb_main" );
    xtsu_copyright_event_remove ( "w_tb_charging" );
    xtsu_copyright_event_remove ( "w_tb_focus" );*/
    

    XtSetArg ( arglist[0], XtNtitle, title );
    XtSetValues ( toplevel_widget, arglist, 1 );

    XtFree ( title );
    title = NULL;
}

void xtsu_copyright_on ( )
{
    caddr_t text;
    
    /*
    **  Fetch current window title, and save it so that we can put it back
    **  later when the copyright is to be removed.
    */
    {
        Arg arglist[1];
        String string;
        
        XtSetArg ( arglist[0], XtNtitle, &string );
        XtGetValues ( toplevel_widget, arglist, 1 );
        title = XtNewString ( string );
    }

    /*
    **  Fetch the copyright notice from the UID file.
    **  
    **  If this doesn't work, do not attempt to place a copyright notice
    **  in the title bar, but just return quietly.
    */
    {
        MrmCode value_type;
        int status;
        
        /* reset the text pointer */
        text = NULL;
        
        /* fetch the literal from the UID file */
        status = MrmFetchLiteral ( s_MrmHierarchy, "copyright_notice",
            display, &text, &value_type );
    
        /* if the fetch failed, report it to standard output */
        if ( status != MrmSUCCESS )
        {
            printf ( "xtsu_copyright_on: can't get copyright notice from UID file\n" );
            if ( status == MrmNOT_FOUND )
            {
                printf ( "xtsu_copyright_on: literal not found\n" );
            }
            return;
        }
    }

    /*
    **  Set the current window title to the copyright notice text.
    */
    {
        Arg arglist[1];
        XtSetArg ( arglist[0], XtNtitle, text );
        XtSetValues ( toplevel_widget, arglist, 1 );
    }
    
    XtFree ( text );

    /* add event handlers */
    {
        xtsu_copyright_event_add ( "w_mb_main" );
/*        xtsu_copyright_event_add ( "w_lb_main" );
        xtsu_copyright_event_add ( "w_tb_charging" );
        xtsu_copyright_event_add ( "w_tb_focus" );*/
    }
}

